/***************************************************************************
 *   This file is part of Kate search plugin                               *
 *   SPDX-FileCopyrightText: 2011 Kåre Särs <kare.sars@iki.fi>                           *
 *                                                                         *
 *   SPDX-License-Identifier: LGPL-2.0-or-later
 ***************************************************************************/

#ifndef HTML_DELEGATE_H
#define HTML_DELEGATE_H

#include <QStyledItemDelegate>

class SPHtmlDelegate : public QStyledItemDelegate
{
public:
    explicit SPHtmlDelegate(QObject *parent);
    ~SPHtmlDelegate() override;

    void paint(QPainter *, const QStyleOptionViewItem &, const QModelIndex &) const override;
    QSize sizeHint(const QStyleOptionViewItem &option, const QModelIndex &index) const override;
};

#endif
